#include "parser.h"

namespace json {
    ArrayParseCall::ArrayParseCall(std::vector<JSON> &result): result(result) {}

    std::unique_ptr<ParsingCall> ArrayParseCall::here(ParserContext& pctx) {
        skipWhitespaces(pctx);
        if (peep(pctx) == ']') {
            skip(pctx);
            return NULL;
        }
        if (got_one) {
            demandSkip(pctx, ',');
        }
        result.emplace_back();
        got_one = true;
        return std::make_unique<ValueParseCall>(result.back());
    }

    DictionaryParseCall::DictionaryParseCall(std::map<std::string, JSON>& result): result(result) {}

    std::unique_ptr<ParsingCall> DictionaryParseCall::here(ParserContext& pctx) {
        skipWhitespaces(pctx);
        if (peep(pctx) == '}') {
            skip(pctx);
            return NULL;
        }
        if (got_one) {
            demandSkip(pctx, ',');
        }
        std::string key = demandStringJson(pctx);
        skipWhitespaces(pctx);
        demandSkip(pctx, ':');
        if (result.count(key) > 0)
            throw bad_syntax();
        got_one = true;
        return std::make_unique<ValueParseCall>(result[key]);
    }
}
