#include "jsonobj.h"
#include "inner_storage.h"
#include <assert.h>

namespace json {
    ArrayData::~ArrayData() {
        for (auto& p: data)
            assert(p.isNull());
    }

    DictionaryData::~DictionaryData() {
        for (auto& p: data) {
            assert(p.second.isNull());
        }
    }

    void stomp(JSON& finished) noexcept {
        finished.value = NULL;
        finished.type = null_symbol;
    }

    void nullify_childless(JSON& obj) noexcept {
        if (obj.type == integer) {
            delete static_cast<Integer*>(obj.value);
        } else if (obj.type == string) {
            delete static_cast<std::string*>(obj.value);
        }
        stomp(obj);
    }

    void setup_nr_iterators(const JSON& obj_child, void* wayback) noexcept {
        if (obj_child.type == array) {
            ArrayData& ad = *static_cast<ArrayData*>(obj_child.value);
            ad.wayback = wayback;
            ad.it = ad.data.begin();
        } else if (obj_child.type == dictionary) {
            DictionaryData& dd = *static_cast<DictionaryData*>(obj_child.value);
            dd.wayback = wayback;
            dd.it = dd.data.begin();
        } else
            assert(false);
    }

    void nullify(JSON &obj) noexcept {
        if (!obj.isNatalistic()) {
            nullify_childless(obj);
            return;
        }
        JSON* current = &obj;
        setup_nr_iterators(obj, NULL);
        while (current) {
            if (current->type == array) {
                ArrayData& ad = *static_cast<ArrayData*>(current->value);
            it_check:
                if (ad.it == ad.data.end()) {
                    stomp(*current);
                    current = static_cast<JSON*>(ad.wayback);
                    delete &ad;
                } else {
                    JSON& child = *ad.it;
                    ++ad.it;
                    if (!child.isNatalistic()) {
                        nullify_childless(child);
                        goto it_check;
                    }
                    setup_nr_iterators(child, current);
                    current = &child;
                }
            } else if (current->type == dictionary) {
                DictionaryData& dd = *static_cast<DictionaryData*>(current->value);
            it_check2:
                if (dd.it == dd.data.end()) {
                    stomp(*current);
                    current = static_cast<JSON*>(dd.wayback);
                    delete &dd;
                } else {
                    JSON& child = dd.it->second;
                    ++dd.it;
                    if (!child.isNatalistic()) {
                        nullify_childless(child);
                        goto it_check2;
                    }
                    setup_nr_iterators(child, current);
                    current = &child;
                }
            }
        }
    }

    /* Strong exception guarantee */
    void copy_childless(JSON& destination, const JSON& source) {
        assert(destination.type == null_symbol);
        if (source.type == integer) {
            destination.value = new Integer(source.asInteger());
        } else if (source.type == string) {
            destination.value = new std::string(source.asString());
        }
        destination.type = source.type;
    }

    /* Basic exception guarantee */
    void setup_it_plus_double_wayback_at_copying(JSON& dest_child, const JSON& src_child, JSON* dest_wayback, const JSON* src_wayback)
    {
        assert(dest_child.isNull());
        if (src_child.type == array) {
            ArrayData& SRC_DATA = *static_cast<ArrayData*>(src_child.value);
            dest_child.value = new ArrayData();
            dest_child.type = array;
            /* Destination is in correct state */
            ArrayData& DEST_DATA = *static_cast<ArrayData*>(dest_child.value);
            DEST_DATA.data.resize(SRC_DATA.data.size());
            DEST_DATA.it = DEST_DATA.data.begin();
            DEST_DATA.wayback = dest_wayback;
            SRC_DATA.it = SRC_DATA.data.begin();
            SRC_DATA.wayback = const_cast<void*>(static_cast<const void*>(src_wayback));
        } else if (src_child.type == dictionary) {
            DictionaryData& SRC_DATA = *static_cast<DictionaryData*>(src_child.value);
            dest_child.value = new DictionaryData();
            dest_child.type = dictionary;
            DictionaryData& DEST_DATA = *static_cast<DictionaryData*>(dest_child.value);
            // Here keeping destination data iterator is unnecessary
            DEST_DATA.wayback = dest_wayback;
            SRC_DATA.it = SRC_DATA.data.begin();
            SRC_DATA.wayback = const_cast<void*>(static_cast<const void*>(src_wayback));
        } else
            assert(false);
    }

    void copy_json(JSON& destination, const JSON& source) {
        assert(destination.type == null_symbol);
        if (!source.isNatalistic()) {
            copy_childless(destination, source);
            return;
        }
        setup_it_plus_double_wayback_at_copying(destination, source, NULL, NULL);
        JSON* dest_cur = &destination;
        const JSON* src_cur = &source;
        while (dest_cur) {
            assert(src_cur);
            if (src_cur->type == array) {
                ArrayData& ad_dest = *static_cast<ArrayData*>(dest_cur->value);
                ArrayData& ad_src = *static_cast<ArrayData*>(src_cur->value);
                it_check:
                if (ad_src.it == ad_src.data.end()) {
                    dest_cur = static_cast<JSON*>(ad_dest.wayback);
                    src_cur = static_cast<const JSON*>(ad_src.wayback);
                } else {
                    JSON& src_child = *ad_src.it;
                    JSON& dest_child = *ad_dest.it;
                    ++ad_src.it;
                    ++ad_dest.it;
                    if (!src_child.isNatalistic()) {
                        copy_childless(dest_child, src_child);
                        goto it_check;
                    }
                    setup_it_plus_double_wayback_at_copying(dest_child, src_child, dest_cur, src_cur);
                    dest_cur = &dest_child;
                    src_cur = &src_child;
                }
            } else if (src_cur->type == dictionary) {
                DictionaryData& dd_dest = *static_cast<DictionaryData*>(dest_cur->value);
                DictionaryData& dd_src = *static_cast<DictionaryData*>(src_cur->value);
                it_check2:
                if (dd_src.it == dd_src.data.end()) {
                    dest_cur = static_cast<JSON*>(dd_dest.wayback);
                    src_cur = static_cast<const JSON*>(dd_src.wayback);
                } else {
                    JSON& src_child = dd_src.it->second;
                    JSON& dest_child = dd_dest.data[dd_src.it->first];  // This function throws
                    ++dd_src.it;
                    if (!src_child.isNatalistic()) {
                        copy_childless(dest_child, src_child);
                        goto it_check2;
                    }
                    setup_it_plus_double_wayback_at_copying(dest_child, src_child, dest_cur, src_cur);
                    dest_cur = &dest_child;
                    src_cur = &src_child;
                }
            }
        }
        assert(!src_cur);
    }

    /* returns false if immediate unequality is obvious, true if both are arrays, both are dicts, or A == B */
    bool eq_compare_simple(const JSON &A, const JSON &B) {
        json_t t = A.type;
        if (t != B.type)
            return false;
        if (t == integer) {
            if (A.asInteger() != B.asInteger())
                return false;
        }
        if (t == string) {
            if (A.asString() != B.asString())
                return false;
        }
        return true;
    }

    template<typename DataT>
    void setup_it_plus_double_wayback_at_cmp_half(
        const JSON& A_child, const JSON& B_child, const JSON* A_wayback, const JSON* B_wayback) {
        DataT& dt_A = *static_cast<DataT*>(A_child.value);
        DataT& dt_B = *static_cast<DataT*>(B_child.value);
        dt_A.it = dt_A.data.begin();
        dt_B.it = dt_B.data.begin();
        dt_A.wayback = (void*)A_wayback;
        dt_B.wayback = (void*)B_wayback;
    }

    void setup_it_plus_double_wayback_at_cmp(const JSON& A_child, const JSON& B_child, const JSON* A_wayback, const JSON* B_wayback) {
        assert(A_child.type == B_child.type);
        if (A_child.type == array) {
            setup_it_plus_double_wayback_at_cmp_half<ArrayData>(A_child, B_child, A_wayback, B_wayback);
        } else if (A_child.type == dictionary) {
            setup_it_plus_double_wayback_at_cmp_half<DictionaryData>(A_child, B_child, A_wayback, B_wayback);
        } else
            assert(false);
    }

    bool eq_compare_json(const JSON &A, const JSON &B) noexcept{
        if (!eq_compare_simple(A, B))
            return false;
        if (!A.isNatalistic())
            return true;
        setup_it_plus_double_wayback_at_cmp(A, B, NULL, NULL);
        const JSON* cur_a = &A;
        const JSON* cur_b = &B;
        while (cur_a) {
            assert(cur_b);
            if (cur_a->type == array) {
                ArrayData& dt_a = *static_cast<ArrayData*>(cur_a->value);
                ArrayData& dt_b = *static_cast<ArrayData*>(cur_b->value);
                again:
                if (dt_a.it == dt_a.data.end()) {
                    if (dt_b.it != dt_b.data.end())
                        return false;
                    cur_a = static_cast<const JSON*>(dt_a.wayback);
                    cur_b = static_cast<const JSON*>(dt_b.wayback);
                } else {
                    if (dt_b.it == dt_b.data.end())
                        return false;
                    JSON& child_A = *(dt_a.it++);
                    JSON& child_B = *(dt_b.it++);
                    if (!eq_compare_simple(child_A, child_B))
                        return false;
                    if (!child_A.isNatalistic())
                        goto again;
                    setup_it_plus_double_wayback_at_cmp(child_A, child_B, cur_a, cur_b);
                    cur_a = &child_A;
                    cur_b = &child_B;
                }
            } else if (cur_a->type == dictionary) {
                DictionaryData& dt_a = *static_cast<DictionaryData*>(cur_a->value);
                DictionaryData& dt_b = *static_cast<DictionaryData*>(cur_b->value);
                again2:
                if (dt_a.it == dt_a.data.end()) {
                    if (dt_b.it != dt_b.data.end())
                        return false;
                    cur_a = static_cast<const JSON*>(dt_a.wayback);
                    cur_b = static_cast<const JSON*>(dt_b.wayback);
                } else {
                    if (dt_b.it == dt_b.data.end())
                        return false;
                    if (dt_a.it->first != dt_b.it->first)
                        return false;
                    JSON& child_A = (dt_a.it++)->second;
                    JSON& child_B = (dt_b.it++)->second;
                    if (!eq_compare_simple(child_A, child_B))
                        return false;
                    if (!child_A.isNatalistic())
                        goto again2;
                    setup_it_plus_double_wayback_at_cmp(child_A, child_B, cur_a, cur_b);
                    cur_a = &child_A;
                    cur_b = &child_B;
                }
            }
        }
        assert(!cur_b);
        return true;
    }

    bool is_subtree_of(const JSON &A, const JSON &B) noexcept {
        if (!B.isNatalistic())
            return &A == &B;
        const JSON* current = &B;
        setup_nr_iterators(B, NULL);
        while (current) {
            if (current == &A)
                return true;
            if (current->type == array) {
                ArrayData& ad = *static_cast<ArrayData*>(current->value);
                it_check:
                if (ad.it == ad.data.end()) {
                    current = static_cast<JSON*>(ad.wayback);
                } else {
                    JSON& child = *(ad.it++);
                    if (!child.isNatalistic()){
                        if (&A == &child)
                            return true;
                        goto it_check;
                    }
                    setup_nr_iterators(child, (void*)current);
                    current = &child;
                }
            } else if (current->type == dictionary) {
                DictionaryData& ad = *static_cast<DictionaryData*>(current->value);
                it_check2:
                if (ad.it == ad.data.end()) {
                    current = static_cast<JSON*>(ad.wayback);
                } else {
                    JSON& child = (ad.it++)->second;
                    if (!child.isNatalistic()){
                        if (&A == &child)
                            return true;
                        goto it_check2;
                    }
                    setup_nr_iterators(child, (void*)current);
                    current = &child;
                }
            }
        }
        return false;
    }
}
