#include "parser.h"
#include <assert.h>

namespace json {
    ParserContext::ParserContext(const std::string& text): text(text)  {}

    bool isEof(ParserContext &pctx) {
        return pctx.pos >= pctx.text.size();
    }

    int peep(ParserContext &pctx) {
        if (pctx.pos >= pctx.text.size())
            return endOfFile;
        return pctx.text[pctx.pos];
    }

    int skip(ParserContext &pctx) {
        if (isEof(pctx))
            throw bad_syntax();
        int v = pctx.text[pctx.pos++];
        if (v == '\n') {
            pctx.line++;
            pctx.column = 0;
        } else
            pctx.column++;
        return v;

    }

    void demandSkip(ParserContext &pctx, char ch) {
        if (peep(pctx) != ch)
            throw bad_syntax();
        skip(pctx);
    }

    bool isWhitespace(int vch) {
        return vch == ' ' || vch == '\t' || vch == '\n';
    }

    void skipWhitespaces(ParserContext &pctx) {
        while (isWhitespace(peep(pctx))) {
            skip(pctx);
        }
    }
}
